# Copyright (C) 2004, 2005 Aaron Bentley
# <aaron.bentley@utoronto.ca>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
import cmd
import sys
import os
import terminal
import readline
import string
from bzrlib.errors import BzrError
from bzrlib.commands import get_cmd_object, get_all_cmds

class PromptCmd(cmd.Cmd):
    def __init__(self):
        cmd.Cmd.__init__(self)
        self.prompt = "bzr> "
        try:
            self.tree = arch.tree_root(".")
        except:
            self.tree = None
        self.set_title()
        self.set_prompt()
        self.identchars += '-'
        self.history_file = os.path.expanduser("~/.bazaar/shell-history")
        readline.set_completer_delims(string.whitespace)
        if os.access(self.history_file, os.R_OK) and \
            os.path.isfile(self.history_file):
            readline.read_history_file(self.history_file)
        self.cwd = os.getcwd()

    def write_history(self):
        readline.write_history_file(self.history_file)

    def do_quit(self, args):
        self.write_history()
        sys.exit(0)

    def do_exit(self, args):
        self.do_quit(args)

    def do_EOF(self, args):
        print
        self.do_quit(args)

    def postcmd(self, line, bar):
        self.set_title()
        self.set_prompt()

    def set_prompt(self):
        if self.tree is not None:
            try:
                prompt = pylon.alias_or_version(self.tree.tree_version, 
                                                self.tree, 
                                                full=False)
                if prompt is not None:
                    prompt = " " + prompt +":"+ pylon.tree_cwd(self.tree)
            except:
                prompt = ""
        else:
            prompt = ""
        self.prompt = "bzr%s> " % prompt

    def set_title(self, command=None):
        try:
            version = pylon.alias_or_version(self.tree.tree_version, self.tree, 
                                             full=False)
        except:
            version = "[no version]"
        if command is None:
            command = ""
        sys.stdout.write(terminal.term_title("bzr %s %s" % (command, version)))

    def do_cd(self, line):
        if line == "":
            line = "~"
        line = os.path.expanduser(line)
        if os.path.isabs(line):
            newcwd = line
        else:
            newcwd = self.cwd+'/'+line
        newcwd = os.path.normpath(newcwd)
        try:
            os.chdir(newcwd)
            self.cwd = newcwd
        except Exception, e:
            print e
        try:
            self.tree = arch.tree_root(".")
        except:
            self.tree = None

    def do_help(self, line):
        Help()(line)

    def default(self, line):
        args = line.split()
        commandname = args.pop(0)
        try:
            cmd_obj = get_cmd_object(commandname)
        except BzrError:
            return os.system(line)


        try:
            return (cmd_obj.run_argv(args) or 0)
        except BzrError, e:
            print e
        except KeyboardInterrupt, e:
            print "Interrupted"
        except Exception, e:
#            print "Unhandled error:\n%s" % errors.exception_str(e)
            print "Unhandled error:\n%s" % (e)


    def completenames(self, text, line, begidx, endidx):
        completions = []
        iter = iter_command_names()
        try:
            if len(line) > 0:
                arg = line.split()[-1]
            else:
                arg = ""
            iter = list(iter_munged_completions(iter, arg, text))
        except Exception, e:
            print e, type(e)
        return list(iter)

    def completedefault(self, text, line, begidx, endidx):
        """Perform completion for native commands.
        
        :param text: The text to complete
        :type text: str
        :param line: The entire line to complete
        :type line: str
        :param begidx: The start of the text in the line
        :type begidx: int
        :param endidx: The end of the text in the line
        :type endidx: int
        """
        try:
            (cmd, args, foo) = self.parseline(line)
            command_obj=find_command(cmd)
            if command_obj is not None:
                return command_obj.complete(args.split(), text)
            elif not self.fake_aba.is_command(cmd) and \
                cmdutil.is_tla_command(cmd):
                iter = cmdutil.iter_supported_switches(cmd)
                if len(args) > 0:
                    arg = args.split()[-1]
                else:
                    arg = ""
                if arg.startswith("-"):
                    return list(cmdutil.iter_munged_completions(iter, arg, 
                                                                text))
                else:
                    return list(cmdutil.iter_munged_completions(
                        cmdutil.iter_file_completions(arg), arg, text))


            elif cmd == "cd":
                if len(args) > 0:
                    arg = args.split()[-1]
                else:
                    arg = ""
                iter = cmdutil.iter_dir_completions(arg)
                iter = cmdutil.iter_munged_completions(iter, arg, text)
                return list(iter)
            elif len(args)>0:
                arg = args.split()[-1]
                iter = cmdutil.iter_file_completions(arg)
                return list(cmdutil.iter_munged_completions(iter, arg, text))
            else:
                return self.completenames(text, line, begidx, endidx)
        except Exception, e:
            print e

def run_shell():
    prompt = PromptCmd()
    try:
        prompt.cmdloop()
    finally:
        prompt.write_history()

def iter_file_completions(arg, only_dirs = False):
    """Generate an iterator that iterates through filename completions.

    :param arg: The filename fragment to match
    :type arg: str
    :param only_dirs: If true, match only directories
    :type only_dirs: bool
    """
    cwd = os.getcwd()
    if cwd != "/":
        extras = [".", ".."]
    else:
        extras = []
    (dir, file) = os.path.split(arg)
    if dir != "":
        listingdir = os.path.expanduser(dir)
    else:
        listingdir = cwd
    for file in iter_combine([os.listdir(listingdir), extras]):
        if dir != "":
            userfile = dir+'/'+file
        else:
            userfile = file
        if userfile.startswith(arg):
            if os.path.isdir(listingdir+'/'+file):
                userfile+='/'
                yield userfile
            elif not only_dirs:
                yield userfile


def iter_dir_completions(arg):
    """Generate an iterator that iterates through directory name completions.

    :param arg: The directory name fragment to match
    :type arg: str
    """
    return iter_file_completions(arg, True)

def iter_command_names(hidden=False):
    for real_cmd_name, cmd_class in get_all_cmds():
        if not hidden and cmd_class.hidden:
            continue
        for name in [real_cmd_name] + cmd_class.aliases:
            yield name

def iter_munged_completions(iter, arg, text):
    for completion in iter:
        completion = str(completion)
        if completion.startswith(arg):
            yield completion[len(arg)-len(text):]

